@@ -1,6 +1,6 @@
 /****************************************************************************
 **
-** Copyright (C) 2016 The Qt Company Ltd.
+** Copyright (C) 2017 The Qt Company Ltd.
 ** Contact: http://www.qt.io/licensing/
 **
 ** This file is part of the Qt Quick Templates 2 module of the Qt Toolkit.
@@ -38,15 +38,14 @@
 #include "qquicktextarea_p_p.h"
 #include "qquickcontrol_p.h"
 #include "qquickcontrol_p_p.h"
+#include "qquickscrollview_p.h"
 
-#include <QtGui/qguiapplication.h>
 #include <QtQml/qqmlinfo.h>
 #include <QtQuick/private/qquickitem_p.h>
-#include <QtQuick/private/qquicktext_p.h>
 #include <QtQuick/private/qquickclipnode_p.h>
 #include <QtQuick/private/qquickflickable_p.h>
 
-#ifndef QT_NO_ACCESSIBILITY
+#if QT_CONFIG(accessibility)
 #include <QtQuick/private/qquickaccessibleattached_p.h>
 #endif
 
@@ -59,7 +58,7 @@ QT_BEGIN_NAMESPACE
     \inqmlmodule QtQuick.Controls
     \since 5.7
     \ingroup qtquickcontrols2-input
-    \brief A multi line text input control.
+    \brief Multi-line text input area.
 
     TextArea is a multi-line text editor. TextArea extends TextEdit with
     a \l {placeholderText}{placeholder text} functionality, and adds decoration.
@@ -79,15 +78,16 @@ QT_BEGIN_NAMESPACE
     decoration of the TextArea scrolls together with the rest of the scrollable
     content.
 
+    \section2 Scrollable TextArea
+
     If you want to make a TextArea scrollable, for example, when it covers
-    an entire application page, attach it to a \l Flickable and combine with a
-    \l ScrollBar or \l ScrollIndicator.
+    an entire application page, it can be placed inside a \l ScrollView.
 
-    \image qtquickcontrols2-textarea-flickable.png
+    \image qtquickcontrols2-textarea-scrollable.png
 
-    \snippet qtquickcontrols2-textarea-flickable.qml 1
+    \snippet qtquickcontrols2-textarea-scrollable.qml 1
 
-    A TextArea that is attached to a Flickable does the following:
+    A TextArea that is placed inside a \l ScrollView does the following:
 
     \list
     \li Sets the content size automatically
@@ -99,24 +99,56 @@ QT_BEGIN_NAMESPACE
 */
 
 /*!
-    \qmlsignal QtQuick.Controls::TextArea::pressAndHold(MouseEvent mouse)
+    \qmlsignal QtQuick.Controls::TextArea::pressAndHold(MouseEvent event)
 
     This signal is emitted when there is a long press (the delay depends on the platform plugin).
-    The \l {MouseEvent}{mouse} parameter provides information about the press, including the x and y
+    The \l {MouseEvent}{event} parameter provides information about the press, including the x and y
     position of the press, and which button is pressed.
+
+    \sa pressed, released
+*/
+
+/*!
+    \qmlsignal QtQuick.Controls::TextArea::pressed(MouseEvent event)
+    \since QtQuick.Controls 2.1 (Qt 5.8)
+
+    This signal is emitted when the text area is pressed by the user.
+    The \l {MouseEvent}{event} parameter provides information about the press,
+    including the x and y position of the press, and which button is pressed.
+
+    \sa released, pressAndHold
+*/
+
+/*!
+    \qmlsignal QtQuick.Controls::TextArea::released(MouseEvent event)
+    \since QtQuick.Controls 2.1 (Qt 5.8)
+
+    This signal is emitted when the text area is released by the user.
+    The \l {MouseEvent}{event} parameter provides information about the release,
+    including the x and y position of the press, and which button is pressed.
+
+    \sa pressed, pressAndHold
 */
 
 QQuickTextAreaPrivate::QQuickTextAreaPrivate()
-    : background(nullptr), focusReason(Qt::OtherFocusReason), accessibleAttached(nullptr), flickable(nullptr)
+    : QQuickTextEditPrivate(),
+#if QT_CONFIG(quicktemplates2_hover)
+      hovered(false),
+      explicitHoverEnabled(false),
+#endif
+      background(nullptr),
+      focusReason(Qt::OtherFocusReason),
+      accessibleAttached(nullptr),
+      flickable(nullptr)
 {
-#ifndef QT_NO_ACCESSIBILITY
+#if QT_CONFIG(accessibility)
     QAccessible::installActivationObserver(this);
 #endif
 }
 
 QQuickTextAreaPrivate::~QQuickTextAreaPrivate()
 {
-#ifndef QT_NO_ACCESSIBILITY
+#if QT_CONFIG(accessibility)
     QAccessible::removeActivationObserver(this);
 #endif
 }
@@ -143,6 +175,52 @@ void QQuickTextAreaPrivate::resizeBackground()
     }
 }
 
+/*!
+    \internal
+
+    Determine which font is implicitly imposed on this control by its ancestors
+    and QGuiApplication::font, resolve this against its own font (attributes from
+    the implicit font are copied over). Then propagate this font to this
+    control's children.
+*/
+void QQuickTextAreaPrivate::resolveFont()
+{
+    Q_Q(QQuickTextArea);
+    inheritFont(QQuickControlPrivate::parentFont(q));
+}
+
+void QQuickTextAreaPrivate::inheritFont(const QFont &f)
+{
+    Q_Q(QQuickTextArea);
+    QFont parentFont = font.resolve(f);
+    parentFont.resolve(font.resolve() | f.resolve());
+
+    const QFont defaultFont = QQuickControlPrivate::themeFont(QPlatformTheme::EditorFont);
+    const QFont resolvedFont = parentFont.resolve(defaultFont);
+
+    const bool changed = resolvedFont != sourceFont;
+    q->QQuickTextEdit::setFont(resolvedFont);
+    if (changed)
+        emit q->fontChanged();
+}
+
+#if QT_CONFIG(quicktemplates2_hover)
+void QQuickTextAreaPrivate::updateHoverEnabled(bool enabled, bool xplicit)
+{
+    Q_Q(QQuickTextArea);
+    if (!xplicit && explicitHoverEnabled)
+        return;
+
+    bool wasEnabled = q->isHoverEnabled();
+    explicitHoverEnabled = xplicit;
+    if (wasEnabled != enabled) {
+        q->setAcceptHoverEvents(enabled);
+        QQuickControlPrivate::updateHoverEnabledRecur(q, enabled);
+        emit q->hoverEnabledChanged();
+    }
+}
+#endif
+
 void QQuickTextAreaPrivate::attachFlickable(QQuickFlickable *item)
 {
     Q_Q(QQuickTextArea);
@@ -158,7 +236,7 @@ void QQuickTextAreaPrivate::attachFlickable(QQuickFlickable *item)
     QObject::connect(flickable, &QQuickFlickable::contentXChanged, q, &QQuickItem::update);
     QObject::connect(flickable, &QQuickFlickable::contentYChanged, q, &QQuickItem::update);
 
-    QQuickItemPrivate::get(flickable)->updateOrAddGeometryChangeListener(this, QQuickItemPrivate::SizeChange);
+    QQuickItemPrivate::get(flickable)->updateOrAddGeometryChangeListener(this, QQuickGeometryChange::Size);
     QObjectPrivate::connect(flickable, &QQuickFlickable::contentWidthChanged, this, &QQuickTextAreaPrivate::resizeFlickableControl);
     QObjectPrivate::connect(flickable, &QQuickFlickable::contentHeightChanged, this, &QQuickTextAreaPrivate::resizeFlickableControl);
 
@@ -178,7 +256,7 @@ void QQuickTextAreaPrivate::detachFlickable()
     QObject::disconnect(flickable, &QQuickFlickable::contentXChanged, q, &QQuickItem::update);
     QObject::disconnect(flickable, &QQuickFlickable::contentYChanged, q, &QQuickItem::update);
 
-    QQuickItemPrivate::get(flickable)->updateOrRemoveGeometryChangeListener(this, QQuickItemPrivate::SizeChange);
+    QQuickItemPrivate::get(flickable)->updateOrRemoveGeometryChangeListener(this, QQuickGeometryChange::Nothing);
     QObjectPrivate::disconnect(flickable, &QQuickFlickable::contentWidthChanged, this, &QQuickTextAreaPrivate::resizeFlickableControl);
     QObjectPrivate::disconnect(flickable, &QQuickFlickable::contentHeightChanged, this, &QQuickTextAreaPrivate::resizeFlickableControl);
 
@@ -245,11 +323,11 @@ void QQuickTextAreaPrivate::resizeFlickableContent()
     flickable->setContentHeight(q->contentHeight() + q->topPadding() + q->bottomPadding());
 }
 
-void QQuickTextAreaPrivate::itemGeometryChanged(QQuickItem *item, const QRectF &newGeometry, const QRectF &oldGeometry)
+void QQuickTextAreaPrivate::itemGeometryChanged(QQuickItem *item, QQuickGeometryChange change, const QRectF &diff)
 {
     Q_UNUSED(item);
-    Q_UNUSED(newGeometry);
-    Q_UNUSED(oldGeometry);
+    Q_UNUSED(change);
+    Q_UNUSED(diff);
 
     resizeFlickableControl();
 }
@@ -278,69 +356,19 @@ void QQuickTextAreaPrivate::implicitHeightChanged()
     emit q->implicitHeightChanged3();
 }
 
-QQuickTextArea::QQuickTextArea(QQuickItem *parent) :
-    QQuickTextEdit(*(new QQuickTextAreaPrivate), parent)
+void QQuickTextAreaPrivate::readOnlyChanged(bool isReadOnly)
 {
-    Q_D(QQuickTextArea);
-    setActiveFocusOnTab(true);
-    d->setImplicitResizeEnabled(false);
-    d->pressHandler.control = this;
-#ifndef QT_NO_CURSOR
-    setCursor(Qt::IBeamCursor);
-#endif
-    QObjectPrivate::connect(this, &QQuickTextEdit::readOnlyChanged,
-                            d, &QQuickTextAreaPrivate::_q_readOnlyChanged);
-}
-
-QQuickTextArea::~QQuickTextArea()
-{
-}
-
-QQuickTextAreaAttached *QQuickTextArea::qmlAttachedProperties(QObject *object)
-{
-    return new QQuickTextAreaAttached(object);
-}
-
-/*!
-    \internal
-
-    Determine which font is implicitly imposed on this control by its ancestors
-    and QGuiApplication::font, resolve this against its own font (attributes from
-    the implicit font are copied over). Then propagate this font to this
-    control's children.
-*/
-void QQuickTextAreaPrivate::resolveFont()
-{
-    Q_Q(QQuickTextArea);
-    inheritFont(QQuickControlPrivate::parentFont(q));
-}
-
-void QQuickTextAreaPrivate::inheritFont(const QFont &f)
-{
-    Q_Q(QQuickTextArea);
-    QFont parentFont = font.resolve(f);
-    parentFont.resolve(font.resolve() | f.resolve());
-
-    const QFont defaultFont = QQuickControlPrivate::themeFont(QPlatformTheme::EditorFont);
-    const QFont resolvedFont = parentFont.resolve(defaultFont);
-
-    const bool changed = resolvedFont != sourceFont;
-    q->QQuickTextEdit::setFont(resolvedFont);
-    if (changed)
-        emit q->fontChanged();
-}
-
-void QQuickTextAreaPrivate::_q_readOnlyChanged(bool isReadOnly)
-{
-#ifndef QT_NO_ACCESSIBILITY
+    Q_UNUSED(isReadOnly);
+#if QT_CONFIG(accessibility)
     if (accessibleAttached)
         accessibleAttached->set_readOnly(isReadOnly);
-#else
-    Q_UNUSED(isReadOnly)
+#endif
+#if QT_CONFIG(cursor)
+    q_func()->setCursor(isReadOnly ? Qt::ArrowCursor : Qt::IBeamCursor);
 #endif
 }
 
-#ifndef QT_NO_ACCESSIBILITY
+#if QT_CONFIG(accessibility)
 void QQuickTextAreaPrivate::accessibilityActiveChanged(bool active)
 {
     if (accessibleAttached || !active)
@@ -363,12 +391,31 @@ QAccessible::Role QQuickTextAreaPrivate::accessibleRole() const
 }
 #endif
 
-void QQuickTextAreaPrivate::deleteDelegate(QObject *delegate)
+QQuickTextArea::QQuickTextArea(QQuickItem *parent)
+    : QQuickTextEdit(*(new QQuickTextAreaPrivate), parent)
 {
-    if (componentComplete)
-        delete delegate;
-    else
-        pendingDeletions.append(delegate);
+    Q_D(QQuickTextArea);
+    setActiveFocusOnTab(true);
+    setAcceptedMouseButtons(Qt::AllButtons);
+    d->setImplicitResizeEnabled(false);
+    d->pressHandler.control = this;
+#if QT_CONFIG(cursor)
+    setCursor(Qt::IBeamCursor);
+#endif
+    QObjectPrivate::connect(this, &QQuickTextEdit::readOnlyChanged,
+                            d, &QQuickTextAreaPrivate::readOnlyChanged);
+}
+
+QQuickTextArea::~QQuickTextArea()
+{
+    Q_D(QQuickTextArea);
+    if (d->flickable)
+        d->detachFlickable();
+}
+
+QQuickTextAreaAttached *QQuickTextArea::qmlAttachedProperties(QObject *object)
+{
+    return new QQuickTextAreaAttached(object);
 }
 
 QFont QQuickTextArea::font() const
@@ -391,9 +438,7 @@ void QQuickTextArea::setFont(const QFont &font)
 
     This property holds the background item.
 
-    \note If the background item has no explicit size specified, it automatically
-          follows the control's size. In most cases, there is no need to specify
-          width or height for a background item.
+    \input qquickcontrol-background.qdocinc notes
 
     \sa {Customizing TextArea}
 */
@@ -409,7 +454,7 @@ void QQuickTextArea::setBackground(QQuickItem *background)
     if (d->background == background)
         return;
 
-    d->deleteDelegate(d->background);
+    QQuickControlPrivate::destroyDelegate(d->background, this);
     d->background = background;
     if (background) {
         background->setParentItem(this);
@@ -440,7 +485,7 @@ void QQuickTextArea::setPlaceholderText(const QString &text)
         return;
 
     d->placeholder = text;
-#ifndef QT_NO_ACCESSIBILITY
+#if QT_CONFIG(accessibility)
     if (d->accessibleAttached)
         d->accessibleAttached->setDescription(text);
 #endif
@@ -450,21 +495,7 @@ void QQuickTextArea::setPlaceholderText(const QString &text)
 /*!
     \qmlproperty enumeration QtQuick.Controls::TextArea::focusReason
 
-    This property holds the reason of the last focus change.
-
-    \note This property does not indicate whether the control has \l {Item::activeFocus}
-          {active focus}, but the reason why the control either gained or lost focus.
-
-    \value Qt.MouseFocusReason         A mouse action occurred.
-    \value Qt.TabFocusReason           The Tab key was pressed.
-    \value Qt.BacktabFocusReason       A Backtab occurred. The input for this may include the Shift or Control keys; e.g. Shift+Tab.
-    \value Qt.ActiveWindowFocusReason  The window system made this window either active or inactive.
-    \value Qt.PopupFocusReason         The application opened/closed a pop-up that grabbed/released the keyboard focus.
-    \value Qt.ShortcutFocusReason      The user typed a label's buddy shortcut
-    \value Qt.MenuBarFocusReason       The menu bar took focus.
-    \value Qt.OtherFocusReason         Another reason, usually application-specific.
-
-    \sa Item::activeFocus
+    \include qquickcontrol-focusreason.qdocinc
 */
 Qt::FocusReason QQuickTextArea::focusReason() const
 {
@@ -490,6 +521,82 @@ bool QQuickTextArea::contains(const QPointF &point) const
     return QQuickTextEdit::contains(point);
 }
 
+/*!
+    \since QtQuick.Controls 2.1 (Qt 5.8)
+    \qmlproperty bool QtQuick.Controls::TextArea::hovered
+    \readonly
+
+    This property holds whether the text area is hovered.
+
+    \sa hoverEnabled
+*/
+bool QQuickTextArea::isHovered() const
+{
+#if QT_CONFIG(quicktemplates2_hover)
+    Q_D(const QQuickTextArea);
+    return d->hovered;
+#else
+    return false;
+#endif
+}
+
+void QQuickTextArea::setHovered(bool hovered)
+{
+#if QT_CONFIG(quicktemplates2_hover)
+    Q_D(QQuickTextArea);
+    if (hovered == d->hovered)
+        return;
+
+    d->hovered = hovered;
+    emit hoveredChanged();
+#else
+    Q_UNUSED(hovered);
+#endif
+}
+
+/*!
+    \since QtQuick.Controls 2.1 (Qt 5.8)
+    \qmlproperty bool QtQuick.Controls::TextArea::hoverEnabled
+
+    This property determines whether the text area accepts hover events. The default value is \c true.
+
+    \sa hovered
+*/
+bool QQuickTextArea::isHoverEnabled() const
+{
+#if QT_CONFIG(quicktemplates2_hover)
+    Q_D(const QQuickTextArea);
+    return d->hoverEnabled;
+#else
+    return false;
+#endif
+}
+
+void QQuickTextArea::setHoverEnabled(bool enabled)
+{
+#if QT_CONFIG(quicktemplates2_hover)
+    Q_D(QQuickTextArea);
+    if (d->explicitHoverEnabled && enabled == d->hoverEnabled)
+        return;
+
+    d->updateHoverEnabled(enabled, true); // explicit=true
+#else
+    Q_UNUSED(enabled);
+#endif
+}
+
+void QQuickTextArea::resetHoverEnabled()
+{
+#if QT_CONFIG(quicktemplates2_hover)
+    Q_D(QQuickTextArea);
+    if (!d->explicitHoverEnabled)
+        return;
+
+    d->explicitHoverEnabled = false;
+    d->updateHoverEnabled(QQuickControlPrivate::calcHoverEnabled(d->parentItem), false); // explicit=false
+#endif
+}
+
 void QQuickTextArea::classBegin()
 {
     Q_D(QQuickTextArea);
@@ -501,21 +608,35 @@ void QQuickTextArea::componentComplete()
 {
     Q_D(QQuickTextArea);
     QQuickTextEdit::componentComplete();
-#ifndef QT_NO_ACCESSIBILITY
+#if QT_CONFIG(quicktemplates2_hover)
+    if (!d->explicitHoverEnabled)
+        setAcceptHoverEvents(QQuickControlPrivate::calcHoverEnabled(d->parentItem));
+#endif
+#if QT_CONFIG(accessibility)
     if (!d->accessibleAttached && QAccessible::isActive())
         d->accessibilityActiveChanged(true);
 #endif
-
-    qDeleteAll(d->pendingDeletions);
-    d->pendingDeletions.clear();
 }
 
 void QQuickTextArea::itemChange(QQuickItem::ItemChange change, const QQuickItem::ItemChangeData &value)
 {
     Q_D(QQuickTextArea);
     QQuickTextEdit::itemChange(change, value);
-    if (change == ItemParentHasChanged && value.item)
+    if ((change == ItemParentHasChanged && value.item) || (change == ItemSceneChange && value.window)) {
         d->resolveFont();
+#if QT_CONFIG(quicktemplates2_hover)
+        if (!d->explicitHoverEnabled)
+            d->updateHoverEnabled(QQuickControlPrivate::calcHoverEnabled(d->parentItem), false); // explicit=false
+#endif
+        if (change == ItemParentHasChanged) {
+            QQuickFlickable *flickable = qobject_cast<QQuickFlickable *>(value.item->parentItem());
+            if (flickable) {
+                QQuickScrollView *scrollView = qobject_cast<QQuickScrollView *>(flickable->parentItem());
+                if (scrollView)
+                    d->attachFlickable(flickable);
+            }
+        }
+    }
 }
 
 void QQuickTextArea::geometryChanged(const QRectF &newGeometry, const QRectF &oldGeometry)
@@ -565,6 +686,24 @@ void QQuickTextArea::focusOutEvent(QFocusEvent *event)
     setFocusReason(event->reason());
 }
 
+#if QT_CONFIG(quicktemplates2_hover)
+void QQuickTextArea::hoverEnterEvent(QHoverEvent *event)
+{
+    Q_D(QQuickTextArea);
+    QQuickTextEdit::hoverEnterEvent(event);
+    setHovered(d->hoverEnabled);
+    event->setAccepted(d->hoverEnabled);
+}
+
+void QQuickTextArea::hoverLeaveEvent(QHoverEvent *event)
+{
+    Q_D(QQuickTextArea);
+    QQuickTextEdit::hoverLeaveEvent(event);
+    setHovered(false);
+    event->setAccepted(d->hoverEnabled);
+}
+#endif
+
 void QQuickTextArea::mousePressEvent(QMouseEvent *event)
 {
     Q_D(QQuickTextArea);
@@ -574,7 +713,12 @@ void QQuickTextArea::mousePressEvent(QMouseEvent *event)
             QQuickTextEdit::mousePressEvent(d->pressHandler.delayedMousePressEvent);
             d->pressHandler.clearDelayedMouseEvent();
         }
+        // Calling the base class implementation will result in QQuickTextControl's
+        // press handler being called, which ignores events that aren't Qt::LeftButton.
+        const bool wasAccepted = event->isAccepted();
         QQuickTextEdit::mousePressEvent(event);
+        if (wasAccepted)
+            event->accept();
     }
 }
 
@@ -617,11 +761,10 @@ void QQuickTextArea::mouseDoubleClickEvent(QMouseEvent *event)
 void QQuickTextArea::timerEvent(QTimerEvent *event)
 {
     Q_D(QQuickTextArea);
-    if (event->timerId() == d->pressHandler.timer.timerId()) {
+    if (event->timerId() == d->pressHandler.timer.timerId())
         d->pressHandler.timerEvent(event);
-    } else {
+    else
         QQuickTextEdit::timerEvent(event);
-    }
 }
 
 class QQuickTextAreaAttachedPrivate : public QObjectPrivate
@@ -632,12 +775,8 @@ public:
     QQuickTextArea *control;
 };
 
-QQuickTextAreaAttached::QQuickTextAreaAttached(QObject *parent) :
-    QObject(*(new QQuickTextAreaAttachedPrivate), parent)
-{
-}
-
-QQuickTextAreaAttached::~QQuickTextAreaAttached()
+QQuickTextAreaAttached::QQuickTextAreaAttached(QObject *parent)
+    : QObject(*(new QQuickTextAreaAttachedPrivate), parent)
 {
 }
 
@@ -646,9 +785,7 @@ QQuickTextAreaAttached::~QQuickTextAreaAttached()
 
     This property attaches a text area to a \l Flickable.
 
-    \snippet qtquickcontrols2-textarea-flickable.qml 1
-
-    \sa ScrollBar, ScrollIndicator
+    \sa ScrollBar, ScrollIndicator, {Scrollable TextArea}
 */
 QQuickTextArea *QQuickTextAreaAttached::flickable() const
 {
@@ -661,7 +798,7 @@ void QQuickTextAreaAttached::setFlickable(QQuickTextArea *control)
     Q_D(QQuickTextAreaAttached);
     QQuickFlickable *flickable = qobject_cast<QQuickFlickable *>(parent());
     if (!flickable) {
-        qmlInfo(parent()) << "TextArea must be attached to a Flickable";
+        qmlWarning(parent()) << "TextArea must be attached to a Flickable";
         return;
     }
 
